##############################################################################
#
#    Analytic Structure, an Odoo module
#    Copyright © 2013, 2022 XCG Consulting (https://xcg-consulting.fr/)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

from odoo import api, fields, models
from odoo.tools import config, frozendict

from .models.analytic_dimension import get_analytic_size


class AddMethod:
    """Utility decorator to add methods to an object or a class."""

    def __init__(self, obj):
        self.obj = obj

    def __call__(self, func):
        setattr(self.obj, func.__name__, func)
        return func


class MetaAnalytic(models.MetaModel):
    """Allow the model to use the classes of the analytic_structure module
    in a more streamlined way.

    The metaclass' behavior is specified by adding the following attributes:

    * _analytic: define the analytic structures to be used by the model.
    * _dimension: bind an analytic dimension to the model.

    A description of the syntax expected for each attribute is available in
    the README file.

    Notes:
    * This metaclass may directly modify attributes that are used by Odoo,
    specifically _inherits and the fields of the class.
    * New superclasses are used to define or override methods, in order to
    avoid interacting with the model's own method (re)definitions.
    """

    def __new__(cls, name, bases, nmspc):
        analytic = nmspc.get("_analytic", {})
        para = nmspc.get("_para_analytic", {})
        dimension = nmspc.get("_dimension", {})

        defaults = nmspc.get("_defaults", None)
        if defaults is None:
            defaults = {}
            nmspc["_defaults"] = defaults

        orm_name = nmspc.get("_name", None)
        if orm_name is None:
            orm_name = nmspc.get("_inherit")

        # Analytic fields should be defined in the _analytic attribute.
        if analytic or para:
            bases = cls._setup_analytic_fields(
                analytic, para, defaults, orm_name, name, bases, nmspc
            )

        # The bound dimension should be defined in the _dimension attribute.
        if dimension:
            bases = cls._setup_bound_dimension(
                dimension, defaults, orm_name, name, bases, nmspc
            )

        return super().__new__(cls, name, bases, nmspc)

    def __init__(cls, name, bases, nmspc):
        super().__init__(name, bases, nmspc)

    @classmethod
    def _setup_analytic_fields(
        cls, analytic, para, defaults, orm_name, name, bases, nmspc
    ):
        """Generate analytic and para-analytic fields on the model."""

        # If _analytic uses a shortcut, convert it into a prefix-model mapping.
        if analytic is True:
            analytic = {"a": orm_name.replace(".", "_")}
        elif isinstance(analytic, str):
            analytic = {"a": analytic}
        nmspc["_analytic_name"] = analytic["a"]  # Save the analytic model name.
        view_subfields = nmspc.get("_analytic_view_subfields", tuple())

        # Create a field that will be used for replacement in the view
        if analytic:
            nmspc["analytic_dimensions"] = fields.Char(
                compute=_analytic_dimensions_empty_field, readonly=True
            )

        col_pattern = "{pre}{n}_{suf}"
        size = get_analytic_size()

        # Generate the fields directly into the nmspc.
        all_analytic = []

        for prefix, model_name in analytic.items():
            # Analytic fields
            all_analytic.append((model_name, prefix, "id"))

            for i in range(1, size + 1):
                col_name = col_pattern.format(pre=prefix, n=i, suf="id")
                domain_field = "nd_id.ns{n}_id.model_name".format(n=i)
                nmspc[col_name] = fields.Many2one(
                    comodel_name="analytic.code",
                    domain=[
                        (domain_field, "=", model_name),
                        ("view_type", "=", False),
                        ("disabled_per_company", "=", False),
                    ],
                    ondelete="restrict",
                    tracking=True,
                    index=True,
                )

        for key, value in para.items():
            # Para-analytic fields
            prefix, suffix = key
            model_name = value["model"]
            all_analytic.append((model_name, prefix, suffix))
            if suffix == "id":
                raise ValueError("Para-analytic suffix cannot be 'id'")

            field_type = value["type"]
            args = value["args"]
            kwargs = value["kwargs"]
            for i in range(1, size + 1):
                col_name = col_pattern.format(pre=prefix, n=i, suf=suffix)
                nmspc[col_name] = field_type(*args, **kwargs)
                if "default" in value:
                    defaults[col_name] = value["default"]

        # In order to preserve inheritance and possible overrides, work on a
        # new class that inherits the given bases, then make our model class
        # inherit from this class.
        superclass_name = "_{name}_SuperAnalytic".format(name=name)
        # Set _register to False in order to prevent its instantiation.
        superclass = type(superclass_name, bases, {"_register": False})

        @AddMethod(superclass)
        @api.model
        def fields_get(self, allfields=None, attributes=None):
            """Override field fetcher to include our analytic fields."""
            ret = super(superclass, self).fields_get(
                allfields=allfields, attributes=attributes
            )
            analytic_obj = self.env["analytic.structure"]
            for _model_name, prefix, suffix in all_analytic:
                ret = analytic_obj.analytic_fields_get(self, ret, prefix, suffix)
            return ret

        @AddMethod(superclass)
        @api.model
        def get_view(self, view_id=None, view_type="form", **options):
            """Override view fetcher to include our analytic fields."""
            ret = super(superclass, self).get_view(
                view_id=view_id,
                view_type=view_type,
                **options,
            )
            analytic_obj = self.env["analytic.structure"]
            for _model_name, prefix, suffix in all_analytic:
                ret = analytic_obj.analytic_get_view(self, ret, prefix, suffix)
            for subfield in view_subfields:
                ret = analytic_obj.analytic_get_view(
                    self.env[self._fields[subfield].comodel_name],
                    ret,
                    subfield=subfield,
                )
            return ret

        @AddMethod(superclass)
        @api.model_create_multi
        @api.returns(orm_name, lambda value: value.id)
        def create(self, vals_list):
            """Performs analytic field validation"""
            res = super(superclass, self).create(vals_list)
            # Throws in case of error
            res._validate_analytic_fields(frozendict(analytic))
            return res

        @AddMethod(superclass)
        def write(self, vals):
            """Performs analytic field validation"""
            res = super(superclass, self).write(vals)
            # Throws in case of error
            self._validate_analytic_fields(frozendict(analytic))
            return res

        # Disable pylint unused argument as the method is meant to be redefined in
        # inherited models.
        @AddMethod(superclass)
        def _validate_analytic_fields(self, analytic):  # pylint: disable=W0613
            """Validation function to validate analytic fields.
            The base implementation doesn't actually do anything.
            :param analytic: frozendict, analytic field parameters, such as
                   they would appear in the '_analytic' in the expanded
                   form, ie. as a prefix => model_name mapping.
            :raises: Validation error when applicable.
            """

        @AddMethod(superclass)
        def _valid_field_parameter(self, field, name):
            # allow tracking on analytic models; see also 'mail.thread'
            return name == "tracking" or super(superclass, self)._valid_field_parameter(
                field, name
            )

        return (superclass,)

    @classmethod
    def _setup_bound_dimension(cls, dimension, defaults, orm_name, name, bases, nmspc):
        """Bind a dimension to the model, creating a code for each record."""

        if dimension is True:
            dimension = {}
        elif isinstance(dimension, str):
            dimension = {"name": dimension}

        dimension_name = dimension.get("name", None)
        if dimension_name is None:
            dimension_name = nmspc.get("_description", False) or orm_name

        column = dimension.get("column", "analytic_id")

        ref_module = dimension.get("ref_module", "")

        ref_id = dimension.get("ref_id", None)
        if ref_id is None:
            ref_id = orm_name.replace(".", "_") + "_analytic_dimension_id"

        code_name = dimension.get("code_name", "name")
        code_description = dimension.get("code_description", "description")
        if code_description not in nmspc:
            code_description = code_name  # Easier for auto-dimensions when no descr.

        # To use an inherited, renamed parent field, you have to give its name.
        sync_parent = dimension.get("sync_parent", False)
        if sync_parent is True:
            sync_parent = nmspc.get("_parent_name", "parent_id")

        parent_column = dimension.get("parent_column", column)

        bound_code_is_required = dimension.get("required", True)

        rel_name = dimension.get("rel_name", tuple())
        if rel_name is True:
            rel_name = "Name"
        if isinstance(rel_name, str):
            rel_name = (rel_name, code_name)

        rel_description = dimension.get("rel_description", tuple())
        if rel_description is True:
            rel_description = "Description"
        if isinstance(rel_description, str):
            rel_description = (rel_description, code_description)

        rel_active = dimension.get("rel_active", tuple())
        if rel_active is True:
            rel_active = "Active"
        if isinstance(rel_active, str):
            rel_active = (rel_active, "active")

        rel_view_type = dimension.get("rel_view_type", tuple())
        if rel_view_type is True:
            rel_view_type = "View type"
        if isinstance(rel_view_type, str):
            rel_view_type = (rel_view_type, "view_type")

        rel_disabled_per_company = dimension.get("rel_disabled_per_company", tuple())
        if rel_disabled_per_company is True:
            rel_disabled_per_company = "Disabled in my company"
        if isinstance(rel_disabled_per_company, str):
            rel_disabled_per_company = (
                rel_disabled_per_company,
                "disabled_per_company",
            )

        # By default, only use inherits if we can be sure there is no conflict
        # on the required fields 'name' and 'nd_id'.
        # There can still be conflicts on analytic_code's optional fields.
        use_inherits = dimension.get("use_inherits", None)
        if use_inherits is None:
            use_inherits = not (
                any(field in nmspc for field in ("name", "nd_id"))
                or nmspc.get("_inherits", False)
                or nmspc.get("_inherit", False)
            )

        code_ref_ids = dimension.get("code_ref_ids", False)
        if code_ref_ids is True:
            code_ref_ids = ref_id

        code_ref_module = dimension.get("code_ref_module", "")

        if use_inherits:
            inherits = nmspc.get("_inherits", {})
            inherits["analytic.code"] = column
            nmspc["_inherits"] = inherits

        # Default column for the underlying analytic code.
        if column not in nmspc:
            nmspc[column] = fields.Many2one(
                "analytic.code",
                "Bound Analytic Code",
                required=bound_code_is_required,
                ondelete="restrict",
            )

        rel_cols = [
            cols
            for cols in [
                rel_name + ("name", "Char", True, ""),
                rel_description + ("description", "Char", False, ""),
                rel_active + ("active", "Boolean", False, True),
                rel_view_type + ("view_type", "Boolean", False, False),
            ]
            if len(cols) == 6
        ]

        if rel_cols:
            # NOT a method nor a class member. 'self' is the analytic_code OSV.
            def _record_from_code_id(self):
                """Get the entries to update from the modified codes."""
                obj = self.env.get(orm_name)
                domain = [(column, "in", self.ids)]
                return obj.search(domain)

            for string, model_col, code_col, dtype, req, default in rel_cols:
                nmspc[model_col] = getattr(fields, dtype)(
                    string=string,
                    related=".".join([column, code_col]),
                    relation="analytic.code",
                    required=req,
                    ondelete="restrict",
                    store=True,
                )
                if model_col not in defaults:
                    defaults[model_col] = default

        # In order to preserve inheritance and possible overrides, work on a
        # new class that inherits the given bases, then make our model class
        # inherit from this class.
        superclass_name = "_{name}_SuperDimension".format(name=name)
        # Set _register to False in order to prevent its instantiation.
        superclass = type(superclass_name, bases, {"_register": False})

        @AddMethod(superclass)
        def _setup_complete(self):
            """Load or create the analytic dimension bound to the model."""

            super(superclass, self)._setup_complete()

            data_obj = self.env["ir.model.data"].sudo()
            try:
                # XXX why set the value on self, as it is not available when
                #  using self.env[model_name]?
                data_obj._xmlid_to_res_id(
                    f"{ref_module}.{ref_id}", raise_if_not_found=True
                )
            except ValueError:
                _bound_dimension_id = self.env["analytic.dimension"].create(
                    {
                        "name": dimension_name,
                        "res_model": orm_name,
                        "column_name": column,
                    }
                )
                data_obj._update_xmlids(
                    [
                        {
                            "noupdate": True,
                            "record": _bound_dimension_id,
                            "xml_id": "%s.%s" % (ref_module, ref_id),
                        }
                    ]
                )

        if code_ref_ids:
            prefix = config.get_misc("analytic", "code_ref_prefix", False)

            # This function is called as a method and can be overridden.
            @AddMethod(superclass)
            def _generate_code_ref_id(self):
                data_obj = self.env["ir.model.data"].sudo()
                records = self

                for record in records:
                    code = record[column]
                    code_ref_id_builder = [prefix] if prefix else []
                    if "company_id" in record and record.company_id:
                        code_ref_id_builder.append(record.company_id.code)
                    code_ref_id_builder.append("ANC")
                    code_ref_id_builder.append(code_ref_ids)
                    code_ref_id_builder.append(code.name)

                    vals = {
                        "name": "_".join(code_ref_id_builder),
                        "module": code_ref_module,
                        "model": "analytic.code",
                        "res_id": code.id,
                    }
                    data_obj.create(vals)

        @AddMethod(superclass)
        @api.model
        def get_bound_dimension(self, raise_if_not_found=True):
            """Return the ``analytic.dimension`` record bound to this model."""
            return self.env.ref(
                f"{ref_module}.{ref_id}", raise_if_not_found=raise_if_not_found
            )

        @AddMethod(superclass)
        @api.model
        def _get_bound_dimension_id(self):
            """Internal, used by MetaAnalytic itself. Get ``analytic.dimension`` ID."""
            return self.sudo().get_bound_dimension().id

        @AddMethod(superclass)
        def get_bound_analytics(self, model: str) -> dict:
            """Get the analytic code bound to this model, ready to be fed into the
            specified target model, when the latter contains a corresponding dimension.
            """
            self.ensure_one()
            dim = self.get_bound_dimension(raise_if_not_found=False)
            if not dim:
                return {}  # Bound dimension not created yet.
            model_dim = self.env["analytic.structure"].get_dimensions(model).get(dim.id)
            if not model_dim:
                return {}  # Bound dimension not used by specified target model.
            code = getattr(self, column)
            if not code:
                return {}  # No bound analytic codes built for this record.
            return {f"a{model_dim}_id": code.id}

        @AddMethod(superclass)
        @api.model
        def _create_analytic_code(self, vals, code_vals):
            # Will be set if a new code is created
            new_code = False

            if use_inherits:
                code_vals.update(vals)
            else:
                code_vals["name"] = vals.get(code_name)
                code_vals["description"] = vals.get(code_description)

            if code_vals.get("name"):
                # OpenERP bug: related fields do not work properly on
                # creation.
                for rel in rel_cols:
                    model_col, code_col = rel[1:3]
                    if model_col in vals:
                        code_vals[code_col] = vals[model_col]
                    elif model_col in self._defaults:
                        code_vals[code_col] = self._defaults[model_col]

                # We have to create the code separately, even with
                # inherits.
                code_obj = self.env["analytic.code"]
                code_vals["nd_id"] = self._get_bound_dimension_id()
                # These are behind-the-scenes links so bypass security rules.
                code = code_obj.sudo().create(code_vals)
                vals[column] = code.id
                new_code = code

            return new_code, vals

        @AddMethod(superclass)
        @api.model_create_multi
        @api.returns(orm_name, lambda value: value.id)
        def create(self, vals_list):
            """Create the analytic code."""
            new_records = None
            for vals in vals_list:
                code_vals = {}

                if sync_parent:
                    parent_code = self._get_code_parent(vals)
                    if parent_code is not None and parent_code:
                        code_vals["code_parent_id"] = parent_code.id

                # Will be set if a new code is created
                new_code = False

                # Direct changes to the 'bound analytic code' field are ignored
                # unless the 'force_code_id' context key is passed as True.
                force_code_id = vals.pop(column, False) or self.env.context.get(
                    "force_code_id"
                )
                if force_code_id:
                    new_code = self._force_code(force_code_id, code_vals)
                    vals[column] = force_code_id

                else:
                    new_code, vals = self._create_analytic_code(vals, code_vals)

                new_record = super(superclass, self).create([vals])
                if new_records is None:
                    new_records = new_record
                else:
                    new_records |= new_record

                # update code_description in case it is a computed field
                if (
                    new_code
                    and self._fields.get(code_description) in self.pool.field_computed
                ):
                    new_code.description = getattr(new_record, code_description)

                if not getattr(new_record, column):
                    if sync_parent:
                        parent_code = self._get_code_parent(vals)
                        if parent_code is not None and parent_code:
                            code_vals["code_parent_id"] = parent_code.id

                    new_code, vals = self._create_analytic_code(
                        {
                            field: extract(
                                getattr(new_record, field), field_data["type"]
                            )
                            for field, field_data in new_record.fields_get().items()
                            if field
                            in (list(vals.keys()) + [code_name, code_description])
                            and field in self.env["analytic.code"]._fields
                        },
                        code_vals,
                    )
                    if new_code:
                        super(superclass, new_record).write({column: new_code.id})

                if code_ref_ids:
                    self._generate_code_ref_id(new_record)

                if new_code:
                    # These are behind-the-scenes links so bypass security rules.
                    new_code.sudo().write(
                        {"origin_id": "{},{}".format(self._name, new_record.id)}
                    )

            return new_records

        @AddMethod(superclass)
        def write(self, vals):
            """Update the analytic code's name if it is not inherited,
            and its parent code if parent-child relations are synchronized.
            """
            code_vals = {}
            news = []
            standard_process = False

            if sync_parent:
                parent_code = self._get_code_parent(vals)
                if parent_code is not None and parent_code:
                    code_vals["code_parent_id"] = parent_code.id
                else:
                    parent = getattr(self, sync_parent)
                    if parent:
                        parent_code = getattr(parent, parent_column)
                        if parent_code is not None and parent_code:
                            code_vals["code_parent_id"] = parent_code.id

            # Direct changes to the 'bound analytic code' field are ignored
            # unless the 'force_code_id' context key is passed as True.
            force_code_id = vals.pop(column, False) or self.env.context.get(
                "force_code_id"
            )
            if force_code_id:
                self._force_code(force_code_id, code_vals)
                vals[column] = force_code_id

            elif use_inherits:
                vals.update(code_vals)

            else:
                name_col = rel_name[1] if rel_name else code_name
                description_col = (
                    rel_description[1] if rel_description else code_description
                )
                if name_col in vals:
                    code_vals["name"] = vals[name_col]
                if description_col in vals:
                    code_vals["description"] = vals[description_col]
                standard_process = True

            res = super(superclass, self).write(vals)

            # If updating records with no code, create these.
            if standard_process:
                code_obj = self.env["analytic.code"]

                for rec in self:
                    code = getattr(rec, column)

                    rec_code_vals = code_vals.copy()
                    rec_vals = {}

                    if not rec_code_vals.get("name"):
                        rec_code_vals["name"] = rec.read([name_col])[0][name_col]
                    if not rec_code_vals.get("description"):
                        rec_code_vals["description"] = self.read([description_col])[0][
                            description_col
                        ]

                    if not code and rec_code_vals.get("name"):
                        news.append(rec.id)
                        rec_code_vals["nd_id"] = rec._get_bound_dimension_id()
                        rec_code_vals["origin_id"] = "{},{}".format(self._name, rec.id)
                        # These are behind-the-scenes links so bypass security
                        # rules.
                        rec_vals[column] = code_obj.sudo().create(rec_code_vals).id

                        super(superclass, rec).write(rec_vals)

                    elif rec_code_vals:
                        # These are behind-the-scenes links so bypass security
                        # rules.
                        code.sudo().write(rec_code_vals)

            if code_ref_ids and news is not False:
                for new in news:
                    self._generate_code_ref_id(new)

            return res

        @AddMethod(superclass)
        def unlink(self, **kwargs):
            """When removing this object, remove all associated analytic
            codes referenced by this object.
            Note: the method will fail if the code is referenced by any other
            object due to the RESTRICT constraint. That is the intended
            behavior.
            """

            # Find all related codes
            codes = self.env["analytic.code"]

            for record in self:
                codes |= getattr(record, column)

            res = super(superclass, self).unlink()

            # These are behind-the-scenes links so bypass security rules.
            codes.sudo().unlink(**kwargs)

            return res

        @AddMethod(superclass)
        def _force_code(self, force_code_id, code_vals):
            code_obj = self.env["analytic.code"]

            if not force_code_id:
                raise ValueError(
                    "An analytic code ID MUST be specified if the "
                    "force_code_id key is enabled in the context"
                )
            force_code = code_obj.browse(force_code_id)
            force_code_dim_id = force_code.nd_id.id
            if force_code_dim_id != self._get_bound_dimension_id():
                raise ValueError(
                    "If specified, codes must belong to the bound "
                    "analytic dimension {}".format(dimension_name)
                )
            if code_vals:
                # These are behind-the-scenes links so bypass security rules.
                force_code.sudo().write(code_vals)
            return force_code

        if sync_parent:
            # This function is called as a method and can be overridden.
            @AddMethod(superclass)
            def _get_code_parent(self, vals):
                """If parent_id is in the submitted values, return the analytic
                code of this parent, to be used as the child's code's parent.
                """
                parent_id = vals.get(sync_parent, None)
                if parent_id is not None:
                    if parent_id:
                        parent_model = self.fields_get([sync_parent])[sync_parent].get(
                            "relation"
                        )

                        if parent_model:
                            res = getattr(
                                self.env[parent_model].browse(parent_id),
                                parent_column,
                            )

                        return res if res else False
                    return False
                return None

        return (superclass,)


def extract(value, typ):
    """Returns the ID, if the value is a browse record.
    Returns the IDs, if the value is a browse record list.
    Otherwise, returns the value.
    """

    result = value
    if typ in ["many2many"]:
        res = value.ids
        res.sort()
        result = [(6, 0, list(set(res)))]
    elif typ == "many2one":
        result = value.id
    elif typ == "one2many":
        result = False

    return result


def _analytic_dimensions_empty_field(self):
    """Utility used when including a dummy "analytic_dimensions" field in
    views.
    """
    for record in self:
        record.analytic_dimensions = ""
