##############################################################################
#
#    Analytic Structure, for Odoo
#    Copyright © 2013 XCG Consulting <https://xcg-consulting.fr>
#
#    Analytic Structure is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    Analytic Structure is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with Analytic Structure.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
from odoo import _lt, exceptions, fields, models
from odoo.tools import config

DIMENSION_DUPLICATE_ERROR = _lt("Both {model1} and {model2} reference {dim}")
NO_MODEL_FOR_DIMENSION_ERROR = _lt("No model matches dimension {dim}")


def get_analytic_size():
    """Return analytic size (from config)
    :return: an int (default to 5)
    """
    return int(config.get_misc("analytic", "analytic_size", 5))


class _dimension_meta(models.MetaModel):
    def __new__(cls, name, bases, nmspc):
        size = get_analytic_size()
        for i in range(1, size + 1):
            nmspc["ns{}_id".format(i)] = fields.One2many(
                comodel_name="analytic.structure",
                inverse_name="nd_id",
                domain=[("ordering", "=", i)],
                auto_join=True,
            )
        return super(_dimension_meta, cls).__new__(cls, name, bases, nmspc)


class AnalyticDimension(models.Model, metaclass=_dimension_meta):
    _name = "analytic.dimension"
    _description = "Analytic Dimension"

    name = fields.Char(
        string="Name",
        size=128,
        translate=config.get_misc("analytic", "translate", False),
        required=True,
    )

    nc_ids = fields.One2many(
        comodel_name="analytic.code", inverse_name="nd_id", string="Codes"
    )

    ns_id = fields.One2many(
        comodel_name="analytic.structure",
        inverse_name="nd_id",
        string="Structures",
    )

    res_model = fields.Char(string="Model")
    column_name = fields.Char()

    _sql_constraints = [("unique_name", "unique(name)", "Name must be unique")]

    def sync_analytic_codes_action(self):
        """Create missing analytic codes"""
        for record in self:
            if not record.res_model:
                raise exceptions.ValidationError(
                    str(NO_MODEL_FOR_DIMENSION_ERROR).format(dim=record.name)
                )

            model_obj = self.env[record.res_model]
            # Look for records with missing analytic codes
            missing_code = model_obj.search([(record.column_name, "=", False)])
            # Create codes for those records using 'write' method
            # defined in MetaAnalytic.
            for dim_record in missing_code:
                dim_record.write({})
