##############################################################################
#
#    Analytic Structure, for Odoo
#    Copyright © 2013 XCG Consulting <https://xcg-consulting.fr>
#
#    Analytic Structure is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    Analytic Structure is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with Analytic Structure.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
import json
import logging
import re
from collections import defaultdict

from lxml import etree

from odoo import _, api, exceptions, fields, models, tools
from odoo.tools import config
from odoo.tools.misc import frozendict

_logger = logging.getLogger(__name__)

_DEFAULT_ANALYTIC_PREFIX = "a"

ORDER_SELECTION = [
    (str(n), f"Analysis {n}")
    for n in range(1, int(config.get_misc("analytic", "analytic_size", 5)) + 1)
]


class AnalyticStructure(models.Model):
    _name = "analytic.structure"
    _description = "Analytic Structure"

    @api.constrains("company_id", "model_name", "ordering")
    def _check_unique_ordering_no_company(self):
        columns = ["company_id", "model_name", "ordering"]
        structures = self.read(columns)
        for structure in structures:
            if structure["company_id"]:
                continue  # Already checked by the SQL constraint.
            domain = [
                ("model_name", "=", structure["model_name"]),
                ("ordering", "=", structure["ordering"]),
            ]
            count = self.search_count(domain)
            if count > 1:
                raise exceptions.ValidationError(
                    _(
                        "One dimension per Analysis slot per object when the "
                        "structure is common to all companies."
                    )
                )

    @tools.ormcache()
    def _get_model_name(self):
        """Looks up the list of model names"""
        return [
            (analytic_name, analytic_name)
            for analytic_name in sorted(
                set(model._analytic_name for model in self._get_models_with_analytic())
            )
        ]

    model_name = fields.Selection(
        selection="_get_model_name", string="Object", required=True
    )
    """Not a model name! It can be any value, just that it usually the same as
    a model name."""
    nd_id = fields.Many2one(
        "analytic.dimension",
        "Related Dimension",
        ondelete="restrict",
        index=True,
        required=True,
    )
    ordering = fields.Selection(
        selection=ORDER_SELECTION, string="Analysis slot", required=True
    )
    company_id = fields.Many2one("res.company", "Company", default=False)
    required = fields.Boolean()

    # do not store those two fields, they depend on code
    # it also fails to compute correctly due to other modules not being loaded
    # yet
    res_model = fields.Char("Model", compute="_compute_res_model_prefix")
    prefix = fields.Char("Prefix", compute="_compute_res_model_prefix")

    _sql_constraints = [
        (
            "unique_ordering",
            "unique(company_id,model_name,ordering)",
            "One dimension per Analysis slot per object per company.",
        )
    ]

    @api.depends("model_name")
    def _compute_res_model_prefix(self):
        name_to_model_analytic_slot = self._get_name_to_model_analytic_slot_dict()
        for structure in self:
            if structure.model_name in name_to_model_analytic_slot:
                model_analytic_slot = name_to_model_analytic_slot[structure.model_name]
                structure.res_model = model_analytic_slot[0]
                structure.prefix = model_analytic_slot[1]
            else:
                raise exceptions.UserError(
                    _("Missing _analytic for %s", structure.model_name)
                )

    def name_get(self):
        names = [(record.id, record.model_name) for record in self]
        return names

    def write(self, vals):
        """Redefined to prevent changes to a structure used."""
        # required is not on this list on purpose
        fields_to_consider = ("model_name", "nd_id", "ordering", "company_id")
        # find out which of these fields values has changed
        changed_values = []
        for field_to_consider in fields_to_consider:
            if field_to_consider in vals and any(
                vals[field_to_consider]
                != (
                    value
                    if field_to_consider in ("model_name", "ordering")
                    else value.id
                )
                for value in self.mapped(field_to_consider)
            ):
                changed_values.append(field_to_consider)
        if any(field in vals.keys() for field in changed_values):
            used_analytic_structures = self.env["analytic.structure"]

            for structure in self:
                search_domain = [(structure.field_name(), "!=", False)]
                # some changes of company_id could be fine, but there is no
                # way to find out the company field on models, it is usually
                # company_id, but it is not a mandatory name and there is no
                # way of knowing in the current code.
                # Adding that field on analytic structures to handle that case
                # seems best.
                if self.env[structure.res_model].search_count(search_domain):
                    used_analytic_structures |= structure

            if used_analytic_structures:
                raise exceptions.UserError(
                    _(
                        "The following structure is in use and can not be changed: %s",
                        used_analytic_structures.mapped("model_name"),
                    )
                )

        return super().write(vals)

    @staticmethod
    def format_field_name(ordering, prefix=_DEFAULT_ANALYTIC_PREFIX, suffix="id"):
        """Return an analytic field's name from its slot, prefix and suffix."""
        # copied bellow
        return "{pre}{n}_{suf}".format(pre=prefix, n=ordering, suf=suffix)

    def field_name(self, suffix="id"):
        """Return an analytic field's name from its prefix and suffix."""
        self.ensure_one()
        # copied above
        return "{pre}{n}_{suf}".format(pre=self.prefix, n=self.ordering, suf=suffix)

    @api.model
    def extract_values(
        self,
        source,
        model,
        prefix="a",
        suffix="id",
        dest_model=None,
        dest_prefix=None,
        dest_suffix=None,
    ):
        """For the given source and destination structures, defined by model,
        prefix and suffix, extract the values of all analytic fields from a
        source. Return the ID of those values, mapped with the field attributed
        to their dimension in the destination structure.

        Use this method to copy analytic data from an object to another.
        The source data can be a browse_record from browse or a dict from read.
        The return value is a dictionary. It can be directly passed as the vals
        or defaults attribute for the ORM methods create, write and copy.

        If not specified, the model, prefix and suffix for the destination
        fields are the same as those for the source fields.
        When a destination model is specified, only the values for dimensions
        present in both models will be returned (at their appropriate slot).
        """

        if dest_prefix is None:
            dest_prefix = prefix
        if dest_suffix is None:
            dest_suffix = suffix

        def add_field():
            src = self.format_field_name(slot, prefix, suffix)
            dest = self.format_field_name(dest_slot, dest_prefix, dest_suffix)
            if src in source:
                try:
                    res[dest] = source[src].id
                except AttributeError:
                    res[dest] = source[src]

        res = {}
        src_dim = self.get_ordering_dimensions(model)

        if dest_model is None or dest_model == model:
            for slots in src_dim.values():
                for slot in slots:
                    dest_slot = slot
                    add_field()

        else:
            dest_dim = self.get_ordering_dimensions(dest_model)

            for dimension, slots in src_dim.items():
                if dimension in dest_dim:
                    dest_slots = dest_dim[dimension]
                    slots_number = len(slots)
                    for i, dest_slot in enumerate(dest_slots):  # noqa: B007
                        if i < slots_number:
                            slot = slots[i]
                            add_field()

        return res

    @api.model
    def get_structures(self, model):
        """Return record sets of every analytic structure entry associated with
        the given model.

        :rtype: Odoo "analytic.structure" record set.
        """

        return self.search([("model_name", "=", model)])

    @api.model
    def get_dimensions(self, model):
        """Return a dictionary that contains the identifier (keys) and ordering
        number (values) of the analytic dimensions linked to the given model.
        """

        return {ans.nd_id.id: ans.ordering for ans in self.get_structures(model)}

    @api.model
    def get_ordering_dimensions(self, model):
        """Return a dictionary that contains the identifier (keys) and the list
        of the ordering numbers (values) of the analytic dimensions linked to
        the given model.
        """

        result = defaultdict(list)
        for ans in self.get_structures(model):
            result[ans.nd_id.id].append(ans.ordering)

        return result

    @api.model
    def get_dimensions_names(self, model):
        """Return a dictionary that contains the ordering numbers (keys) and
        names (values) of the analytic dimensions linked to the given model.
        """

        return {ans.ordering: ans.nd_id.name for ans in self.get_structures(model)}

    @api.model
    def analytic_fields_get(self, model, field_get_ret, prefix="a", suffix="id"):
        """Set the label values for the analytic fields."""

        ans_dict = self.get_dimensions_names(model._analytic_name)

        regex = r"{pre}(\d+)_{suf}".format(pre=prefix, suf=suffix)
        match_fct = re.compile(regex).search
        matches = [_f for _f in map(match_fct, field_get_ret.keys()) if _f]

        for match in matches:
            field = match.group(0)
            slot = match.group(1)
            field_get_ret[field]["string"] = ans_dict.get(
                "{0}".format(slot), "{0}{1}".format(prefix.upper(), slot)
            )

        return field_get_ret

    @api.model
    def analytic_get_view(self, model, view, prefix="a", suffix="id", subfield=None):
        """Replace `analytic_dimensions` by actual analytic fields such as `a1_id`."""
        doc = etree.fromstring(view["arch"])

        # Look for any field named 'analytic_dimensions' and with the right prefix.
        name_cond = '@name="analytic_dimensions"'
        if prefix == "a":
            pre_cond = '(@prefix="a" or not(@prefix))'
        else:
            pre_cond = '@prefix="{pre}"'.format(pre=prefix)
        if suffix == "id":
            suf_cond = '(@suffix="id" or not(@suffix))'
        else:
            suf_cond = '@suffix="{suf}"'.format(suf=suffix)
        condition = "{0} and {1} and {2}".format(name_cond, pre_cond, suf_cond)
        xpath = f"//field[{condition}]"
        if subfield:
            xpath = f"//field[@name='{subfield}']" + xpath
        for node in doc.xpath(xpath):
            if not subfield and list(node.iterancestors(tag="field")):
                # This xpath caught an analytic_dimensions field within another field;
                # this must be a sub-view embedded within main view (eg sale.order.line
                # within sale.order views).
                continue
            parent = node.getparent()

            structures = self.get_structures(model._analytic_name)
            if structures:
                a_fields = []
                for structure in structures.sorted(key=lambda it: int(it.ordering)):
                    field = self.format_field_name(structure.ordering, prefix, suffix)
                    a_fields.append((field, structure.required))

                # Declare new fields within tuple in the return dict.
                ret_models = view["models"].copy()  # frozendict➔dict
                ret_models[model._name] = tuple(  # list➔tuple
                    list(ret_models[model._name])  # tuple➔list
                    + list(field for field, _required in a_fields)
                )
                view["models"] = frozendict(ret_models)  # dict➔frozendict

                # Insert fields in the view's architecture where placeholder field was.
                node_pos = parent.index(node)
                attrs_to_copy = node.attrib
                modifiers_to_copy = json.loads(attrs_to_copy.get("modifiers", "{}"))
                for field, required in a_fields:
                    attrs = {
                        attr: value
                        for attr, value in attrs_to_copy.items()
                        if attr not in ("name", "prefix", "suffix", "modifiers")
                    }
                    attrs["name"] = field
                    modifiers = {
                        attr: value
                        for attr, value in modifiers_to_copy.items()
                        if attrs_to_copy.get(attr) in ("True", "true", "1")
                    }
                    if required:
                        modifiers["required"] = True
                    attrs["modifiers"] = json.dumps(modifiers)
                    parent.insert(node_pos, etree.Element("field", attrs))
                    node_pos += 1

            # Always remove placeholder field, even without any defined structures.
            parent.remove(node)

        view["arch"] = etree.tostring(doc, encoding="unicode")
        return view

    @tools.ormcache()
    def _get_models_with_analytic(self):
        """Return models with analytics"""
        registry = self.env.registry
        return [
            model
            for name, model in registry.items()
            if getattr(model, "_analytic", False)
        ]

    @tools.ormcache()
    def _get_name_to_model_analytic_slot_dict(self):
        name_to_model_dict = {}
        for model in self._get_models_with_analytic():
            analytic = model._analytic
            if analytic is True:
                analytic_name = model._name.replace(".", "_")
                name_to_model_dict[analytic_name] = (model._name, "a")
            elif isinstance(analytic, str):
                name_to_model_dict[analytic] = (model._name, "a")
            else:
                # Expecting analytic to be a dict
                name_to_model_dict.update(
                    {value: (model._name, key) for key, value in analytic.items()}
                )
        return name_to_model_dict
